/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 */

/*
 *    EncoderData.java
 *    Copyright (C) 2012 Peng Chen, King Abdullah University of Science and 
 *    Technology, King of Saudi Arabia
 *
 */


import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Scanner;
import java.util.Set;

import javax.rmi.CORBA.Util;

import org.apache.commons.math3.linear.RealMatrix;
import org.apache.commons.math3.stat.correlation.PearsonsCorrelation;
import org.biojava.bio.BioException;
import org.biojava.bio.proteomics.aaindex.AAindex;
import org.biojava.bio.proteomics.aaindex.AAindexStreamReader;
import org.biojava.bio.proteomics.aaindex.SimpleSymbolPropertyTableDB;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.symbol.SymbolList;
import org.biojavax.SimpleNamespace;
import org.biojavax.bio.seq.RichSequence;
import org.biojavax.bio.seq.RichSequenceIterator;

import pchen.classifiers.Normalization;
import weka.core.Instances;
import weka.core.Utils;

public class EncoderData {
	final static String aa1name = "ACDEFGHIKLMNPQRSTVWYXXX";
	private static ArrayList<String> querySeq = new ArrayList<String>();
	private static ArrayList<String> queryName = new ArrayList<String>();
	static int aaDistSeq = 24;

	/**
	 * @param args
	 */
	public EncoderData() {
	}


	/**
	 * Evaluate classifiers for the top efficient feature subsets
	 * 
	 * @param inputSeqAdd
	 *            File address for the input sequences.
	 * @param outputAddr
	 *            Address for storing the output, such as target information,
	 *            sequence.
	 * @return null
	 * @method First parse fasta sequences to individual fasta sequence file,
	 *         obtain contact map and the corresponding contact states, and
	 *         finally parse sequence profile file to a matrix each of which row
	 *         is for a residue representing by 20 values of amino acids.
	 * @throws
	 */
	public EncoderData(String inputSeqAdd, String outputAddr) {
		SimpleNamespace ns = new SimpleNamespace("bj_blast");
		try {
			// read Fasta sequence file
			RichSequenceIterator rs = RichSequence.IOTools.readFastaProtein(
					new BufferedReader(
							new FileReader(inputSeqAdd + "FASTA.seq")), ns);
			while (rs.hasNext()) {
				RichSequence rr = rs.nextRichSequence();
				queryName.add(rr.getName());
				querySeq.add(rr.seqString());
			}
			File dir = new File(inputSeqAdd + "SP\\");
			File[] files = dir.listFiles();
			// treat for all of the files in dataset
			for (int index = 0; index < files.length; index++) {
				if ((new File(outputAddr + queryName.get(index))).exists() == false)
					(new File(outputAddr + queryName.get(index))).mkdir();

				// save individual sequence file in FASTA format
				String title = ">" + queryName.get(index);
				String seqPDB = querySeq.get(index);
				SaveSeqTOFasta(title, seqPDB, outputAddr + queryName.get(index)
						+ "\\Seq.txt");

				// obtain sequence profiles of residue from the BLAST file
				Scanner in = new Scanner(new File(inputSeqAdd + "SP\\"
						+ queryName.get(index) + ".sp"));
				String line = in.nextLine();
				String[] strLine = line.trim().split(" ");
				double[][] seqProfile = new double[seqPDB.length()][20];
				System.out.println(queryName.get(index));
				int numLine = 0;
				while (in.hasNextLine()) { // iterate thru the input source
					line = in.nextLine();
					strLine = line.trim().split(" ");
					for (int i = 0; i < 20; i++) {
						seqProfile[numLine][i] = Double.parseDouble(strLine[i]);
					}
					numLine++;
				}
			}

		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (NoSuchElementException e) {
			e.printStackTrace();
		} catch (BioException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Function to get residue vectors that integrates sequence profile with one
	 * AAindex1 by the use of window technique
	 * 
	 * @param title
	 * 
	 * @param seqPDB
	 *            Sequence string for a protein chain.
	 * @param cmToBinary
	 *            Store target information for a pair of residues, such as 0 1
	 *            45 , 1 1 46, etc.
	 * @param seqProfile
	 *            Sequence profile for each residue in protein chain.
	 * @param winLen
	 *            Window length using window technique.
	 * @param AAindexEach
	 *            Each AAindex score array.
	 * @return Input Input encoder for each pair of residues.
	 * @method First get the production of sequence profile of a residue to the
	 *         corresponding AAindex score, yield the standard deviation of the
	 *         productions over the twenty types of amino acids, the standard
	 *         deviation is then encoded for the residue in protein chain.
	 * @throws
	 */
	public static double[][] EncoderWindow(String seqPDB, double[] target,
			double[][] seqProfile, int winLen, double[] AAindexEach,
			String title) {
		int radius = (int) (winLen - 1) / 2;
		int numInstances = seqPDB.length();
		double[][] spMultiplyAAindex = new double[numInstances][20];
		double[] vector = { 5, 4, 3, 2, 1, 0, 1, 2, 3, 4, 5 };
		double mu = Utils.mean(vector);
		double var = Utils.variance(vector);
		double[] std = new double[numInstances];
		// calculate the standard deviation of spMultiplyAAindex
		for (int i = 0; i < numInstances; i++) {
			for (int j = 0; j < 20; j++)
				spMultiplyAAindex[i][j] = seqProfile[i][j] * AAindexEach[j];
			double mean = 0;
			for (int j = 0; j < 20; j++)
				mean += spMultiplyAAindex[i][j];
			double addSubSqr = 0;
			for (int j = 0; j < 20; j++)
				addSubSqr += Math.pow(spMultiplyAAindex[i][j] - mean, 2);
			std[i] = Math.pow(addSubSqr / 19, 0.5);
		}

		double[][] Input = null;
		// only retain some vectors
		int numInstance = 0;
		for (int i = 0; i < numInstances; i++)
			if (target[i] != 0)
				numInstance++;
		Input = new double[numInstance][winLen];

		if (title == "test") {
			numInstance = 0;
			for (int i = 0; i < numInstances; i++) {
				if (target[i] != 0) {
					if (numInstances < winLen) {
						for (int j = -radius; j <= radius; j++) {
							if ((i + j) < 0 | (i + j) >= numInstances) {
								Input[numInstance][j + radius] = 0;
							} else {
								Input[numInstance][j + i] = std[i + j]
										* (Math.exp(-Math.pow(Math.abs(Math
												.abs(j))
												- mu, 2)
												/ (2 * var)) / (Math.sqrt(2
												* Math.PI * var)));
							}
						}

					} else {
						if (i < winLen)
							for (int j = 0; j < winLen; j++)
								Input[numInstance][j] = std[j]
										* (Math.exp(-Math.pow(Math.abs(Math
												.abs(j - radius))
												- mu, 2)
												/ (2 * var)) / (Math.sqrt(2
												* Math.PI * var)));
						else if (i > numInstances - winLen)
							for (int j = 0; j < winLen; j++)
								Input[numInstance][j] = std[numInstances - 1
										- j]
										* (Math.exp(-Math.pow(Math.abs(Math
												.abs(j - radius))
												- mu, 2)
												/ (2 * var)) / (Math.sqrt(2
												* Math.PI * var)));
						else
							for (int j = -radius; j <= radius; j++)
								Input[numInstance][j + radius] = std[i + j]
										* (Math.exp(-Math.pow(Math.abs(Math
												.abs(j))
												- mu, 2)
												/ (2 * var)) / (Math.sqrt(2
												* Math.PI * var)));
					}
					numInstance++;
				}
			}
		} else {
			numInstance = 0;
			for (int i = 0; i < numInstances; i++) {
				if (target[i] != 0) {
					if (numInstances < winLen) {
						for (int j = -radius; j <= radius; j++) {
							if ((i + j) < 0 | (i + j) >= numInstances) {
								Input[numInstance][j + radius] = 0;
							} else {
								Input[numInstance][j + i] = std[i + j]
										* (Math.exp(-Math.pow(Math.abs(Math
												.abs(j))
												- mu, 2)
												/ (2 * var)) / (Math.sqrt(2
												* Math.PI * var)));
							}
						}
					} else {
						if (i < winLen)
							for (int j = 0; j < winLen; j++)
								Input[numInstance][j] = std[j]
										* (Math.exp(-Math.pow(Math.abs(Math
												.abs(j - radius))
												- mu, 2)
												/ (2 * var)) / (Math.sqrt(2
												* Math.PI * var)));
						else if (i > numInstances - winLen)
							for (int j = 0; j < winLen; j++)
								Input[numInstance][j] = std[numInstances - 1
										- j]
										* (Math.exp(-Math.pow(Math.abs(Math
												.abs(j - radius))
												- mu, 2)
												/ (2 * var)) / (Math.sqrt(2
												* Math.PI * var)));
						else
							for (int j = -radius; j <= radius; j++)
								Input[numInstance][j + radius] = std[i + j]
										* (Math.exp(-Math.pow(Math.abs(Math
												.abs(j))
												- mu, 2)
												/ (2 * var)) / (Math.sqrt(2
												* Math.PI * var)));
					}
					numInstance++;
				}
			}
		}
		Input = Normalization.Scale(Input);
		return Input;
	}

	/**
	 * Function to get targets retained for valid encoding vectors
	 * 
	 * @param target
	 *            Store target information for a residue, such as [0 1 -1 , 1 1
	 *            0 ...].
	 * @return targetRetained Targets retained.
	 * @method Get targets having value 1 or -1, and then change -1 to 0 for the
	 *         non-Hotspot residues.
	 * @throws
	 */
	public static double[] Target(double[] target) {
		int numInstances = target.length;
		int numInstance = 0;
		for (int i = 0; i < target.length; i++)
			if (target[i] != 0)
				numInstance++;

		double[] targetRetained = new double[numInstance];
		numInstance = 0;
		for (int i = 0; i < numInstances; i++) {
			if (target[i] != 0) {
				targetRetained[numInstance] = target[i];
				if (targetRetained[numInstance] == -1)
					targetRetained[numInstance] = 0;
				numInstance++;
			}
		}

		return targetRetained;
	}

	/**
	 * Function to get targets retained for all encoding vectors
	 * 
	 * @param target
	 *            Store target information for a residue, such as [0 1 -1 , 1 1
	 *            0 ...].
	 * @return targetRetained Targets retained.
	 * @method Get targets having value 1 or 0 representing for Hotspot residue
	 *         or non-Hotspot residue.
	 * @throws
	 */
	public static double[] TargetAll(double[] target) {
		int numInstances = target.length;
		double[] target0 = target.clone();
		for (int i = 0; i < numInstances; i++)
			if (target[i] == -1)
				target0[i] = 0;

		return target0;
	}

	/**
	 * Get information from AAindex1 dataset
	 * 
	 * @param AAindex1FileAddr
	 *            Address of the AAindex1 file
	 * @return AAindex1 Whose row denotes index array for 20 types of amino
	 *         acids.
	 * @method Retrieve AAindex1 information from file and save it into a matrix
	 *         each row of which is for 20 types of amino acids.
	 * @throws
	 */
	@SuppressWarnings("unchecked")
	public static double[][] AAindex1Info(String AAindex1FileAddr) {
		SimpleSymbolPropertyTableDB db;
		double[][] AAindex1 = null;
		try {
			// obtain AAindex1 information and save to tables
			db = new SimpleSymbolPropertyTableDB(new AAindexStreamReader(
					new FileReader(AAindex1FileAddr)));
			AAindex1 = new double[db.numTables()][20];
			Set set = db.names();
			int num = 0;
			// parse tables to a matrix which row is for 20 types of amino acids
			for (java.util.Iterator j = set.iterator(); j.hasNext();) {
				String str = (String) j.next();
				// System.out.println(str);
				SymbolList symbols = ProteinTools
						.createProtein("ARNDCEQGHILKMFPSTWYV");
				AAindex hydrophobicity = (AAindex) db.table(str);
				// System.out.println(symbols.length());

				for (int i = 1; i <= symbols.length(); i++) {
					AAindex1[num][i - 1] = hydrophobicity
							.getDoubleValue(symbols.symbolAt(i));
				}
				num++;
			}
		} catch (NullPointerException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (BioException e) {
			e.printStackTrace();
		}

		return AAindex1;
	}

	/**
	 * Get independent AAindex properties. The new version is different with the
	 * old for paper in Proteins that randomly choose an initial property from
	 * the AAindex dataset, compute correlations with others, and remove
	 * properties having correlation above 0.7 with it. Then, in the retained
	 * properties, we randomly choose another initial property and repeated the
	 * above removal step. Step by step, 132 properties that no pair of
	 * properties has a correlation above 0.7 are achieved.
	 * 
	 * @param AAindex1FileAddr
	 *            Address of the AAindex1 file
	 * @return AAindex1 Whose row denotes index array for 20 types of amino
	 *         acids.
	 * @method Retrieve AAindex1 information from file and save it into a matrix
	 *         each row of which is for 20 types of amino acids. First, for each
	 *         property, count the number of properties that having a
	 *         correlation more than 0.7 with the property, rank properties with
	 *         the correlation number by descend, retain property from the first
	 *         one and remove other correlation properties. Finally, an set of
	 *         independent properties is achieved. In addition, remove the
	 *         similar rows from the matrix and the rows retained finally are
	 *         for our requirement.
	 * @throws IOException
	 * @throws BioException
	 * @throws FileNotFoundException
	 * @throws NullPointerException
	 */
	@SuppressWarnings( { "unchecked", "null" })
	public static double[][] getIrrelevanceAAindex2(String aAindex1FileAddr,
			double cutoff) throws NullPointerException, FileNotFoundException,
			BioException, IOException {
		SimpleSymbolPropertyTableDB db;
		double[][] AAindex1 = null;
		double[][] AAindex1NonSimilar = null;

		// obtain AAindex1 information and save to tables
		db = new SimpleSymbolPropertyTableDB(new AAindexStreamReader(
				new FileReader(aAindex1FileAddr)));
		Set set = db.names();
		int num = set.size();
		String[] accessionNumber = new String[num];
		String[] articleTitle = new String[num];

		AAindex1 = new double[num][20];
		double[] NAN = new double[num];
		int numAA = 0;
		for (java.util.Iterator j = set.iterator(); j.hasNext();) {
			String str = (String) j.next();
			SymbolList symbols = ProteinTools
					.createProtein("ARNDCEQGHILKMFPSTWYV");
			AAindex aaindex1 = (AAindex) db.table(str);
			accessionNumber[numAA] = aaindex1.accessionNumber();
			articleTitle[numAA] = aaindex1.getDescription();
			// System.out.println(numAA + 1 + " " + accessionNumber[numAA]
			// + " " + articleTitle[numAA]);
			for (int i = 1; i <= symbols.length(); i++) {
				AAindex1[numAA][i - 1] = aaindex1.getDoubleValue(symbols
						.symbolAt(i));
				if (Double.isNaN(AAindex1[numAA][i - 1])) {
					NAN[numAA] = 1;
					AAindex1[numAA][i - 1] = 0;
				}
			}
			numAA++;
		}
		double[][] AAindex0 = new double[AAindex1[0].length][AAindex1.length];
		for (int i = 0; i < AAindex1.length; i++) {
			for (int j = 0; j < AAindex1[0].length; j++) {
				AAindex0[j][i] = AAindex1[i][j];
			}
		}
		double[] pos2sum = new double[AAindex1.length];
		PearsonsCorrelation corr = (new PearsonsCorrelation());
		RealMatrix m = corr.computeCorrelationMatrix(AAindex0);
		for (int i = 0; i < m.getRowDimension(); i++) {
			for (int j = 0; j < m.getColumnDimension(); j++) {
				if (Math.abs(m.getEntry(i, j)) >= cutoff) {
					pos2sum[i]--;
				}
			}
		}
		int[] sortPos2sum = Utils.sort(pos2sum); // ascending order

		int[] initialStatus = new int[AAindex1.length];
		ArrayList<Integer> pos4NonSimilar = new ArrayList<Integer>();
		for (int i = 0; i < AAindex1.length; i++) {
			if (initialStatus[sortPos2sum[i]] == 0 && NAN[sortPos2sum[i]] == 0) {
				for (int j = i + 1; j < AAindex1.length; j++) {
					if (initialStatus[sortPos2sum[j]] == 0
							&& NAN[sortPos2sum[j]] == 0) {
						if (Math
								.abs(m.getEntry(sortPos2sum[i], sortPos2sum[j])) >= cutoff) {
							initialStatus[sortPos2sum[j]] = 1;
						}
					}
				}
				pos4NonSimilar.add(sortPos2sum[i]);
			}
		}

		AAindex1NonSimilar = new double[pos4NonSimilar.size()][AAindex1[0].length];
		for (int i = 0; i < pos4NonSimilar.size(); i++) {
			int pos = pos4NonSimilar.get(i);
			// System.out.print(pos + 1 + " ");
			AAindex1NonSimilar[i] = AAindex1[pos];
			// display the selected features
			// System.out.println(i + 1 + " " + accessionNumber[pos] + " "
			// + articleTitle[pos]);
		}
		// System.out.println(pos4NonSimilar.size());

		return AAindex1NonSimilar;
	}

	/**
	 * Get independent properties from AAindex dataset itself. The correlation
	 * cutoff is set to 0.8 in the dataset.
	 * 
	 * @param AAindex1FileAddr
	 *            Address of the AAindex1 file
	 * @return AAindex1 Whose row denotes index array for 20 types of amino
	 *         acids.
	 * @method Retrieve AAindex1 information from file and save it into a matrix
	 *         each row of which is for 20 types of amino acids. In addition,
	 *         remove the similar rows from the matrix and the rows retained
	 *         finally are for our requirement.
	 * @throws
	 */
	@SuppressWarnings("unchecked")
	public static double[][] getIrrelevanceAAindex1(String aAindex1FileAddr) {
		SimpleSymbolPropertyTableDB db;
		double[][] AAindex1 = null;
		int[] remainAccess = null;
		try {
			// obtain AAindex1 information and save to tables
			db = new SimpleSymbolPropertyTableDB(new AAindexStreamReader(
					new FileReader(aAindex1FileAddr)));
			Set set = db.names();
			int num = 0;
			String[] accessionNumber = new String[db.numTables()];
			for (java.util.Iterator j = set.iterator(); j.hasNext();) {
				String str = (String) j.next();
				AAindex aaindex1 = (AAindex) db.table(str);
				accessionNumber[num] = aaindex1.accessionNumber();
				num++;
			}

			remainAccess = new int[db.numTables()];
			set = db.names();
			// parse tables to a matrix which row is for 20 types of amino acids
			// and remove the similar array
			for (java.util.Iterator j = set.iterator(); j.hasNext();) {
				String str = (String) j.next();
				// SymbolList symbols = ProteinTools
				// .createProtein("ARNDCEQGHILKMFPSTWYV");
				AAindex aaindex1 = (AAindex) db.table(str);

				Map similarEntry = aaindex1.similarEntries();
				java.util.Iterator it = similarEntry.entrySet().iterator();
				while (it.hasNext()) {
					Map.Entry entry = (Map.Entry) it.next();
					Object key = entry.getKey();
					for (int i = 0; i < db.numTables(); i++) {
						if (key.toString().matches(accessionNumber[i])) {
							remainAccess[i] = 1;
						}
					}
				}

			}
			num = 0;
			for (int i = 0; i < db.numTables(); i++) {
				if (remainAccess[i] == 0)
					num++;
			}
			AAindex1 = new double[num][20];
			num = 0;
			int numIrrelevance = 0;
			db = new SimpleSymbolPropertyTableDB(new AAindexStreamReader(
					new FileReader(aAindex1FileAddr)));
			set = db.names();
			for (java.util.Iterator j = set.iterator(); j.hasNext();) {
				String str = (String) j.next();
				SymbolList symbols = ProteinTools
						.createProtein("ARNDCEQGHILKMFPSTWYV");
				AAindex hydrophobicity = (AAindex) db.table(str);

				if (remainAccess[num] == 0) {
					// System.out.println("'"+hydrophobicity.accessionNumber()+"';");
					for (int i = 1; i <= symbols.length(); i++) {
						AAindex1[numIrrelevance][i - 1] = hydrophobicity
								.getDoubleValue(symbols.symbolAt(i));
					}
					numIrrelevance++;
				}
				num++;
			}
		} catch (NullPointerException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (BioException e) {
			e.printStackTrace();
		}

		return AAindex1;
	}

	/**
	 * Save sequence information in Fasta
	 * 
	 * @param title
	 *            Sequence title
	 * @param seq
	 *            Sequence of amino acid residue in one letter
	 * @param outputSaveAdd
	 *            Address of the saved file
	 * @return null
	 * @method
	 * @throws
	 */
	private static void SaveSeqTOFasta(String title, String seq,
			String outputSaveAdd) {
		try {
			FileOutputStream writer = new FileOutputStream(outputSaveAdd);
			DataOutputStream osw = new DataOutputStream(writer);
			osw.writeBytes(title.toString());
			osw.write("\r\n".getBytes()); // enter
			osw.writeBytes(seq.toString());
			osw.close();
			writer.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Save 2D array into a file in arff format.
	 * 
	 * @param cmToBinary
	 *            A 2D array, which row represents an instance containing target
	 *            value.
	 * @param outputSaveAdd
	 *            Address of the AAindex1 file to be stored
	 * @return null
	 * @method Save 2D array into a file in arff format. Each element has three
	 *         valid numbers.
	 * @throws
	 */
	private static void SaveDataTOFile(int[][] cmToBinary, String outputSaveAdd) {
		try {
			FileOutputStream writer = new FileOutputStream(outputSaveAdd);
			DataOutputStream osw = new DataOutputStream(writer);
			/*
			 * DecimalFormat df = new DecimalFormat("0.000"); // format a double
			 * value String num = df.format(a);
			 */
			DecimalFormat df = new DecimalFormat("0");
			int numPair = cmToBinary.length;
			for (int i = 0; i < numPair; i++) {
				for (int j = 0; j < 3; j++) {
					osw
							.write((String.valueOf(df.format(cmToBinary[i][j])) + " ")
									.getBytes());
				}
				osw.write("\r\n".getBytes()); // enter
			}
			osw.close();
			writer.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

}
