/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 */

/*
 *    Predictor.java
 *    Copyright (C) 2012 Peng Chen, King Abdullah University of Science and 
 *    Technology, King of Saudi Arabia
 *
 */

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.NoSuchElementException;
import java.util.Random;

import org.biojava.bio.BioException;
import org.biojavax.SimpleNamespace;
import org.biojavax.bio.seq.RichSequence;
import org.biojavax.bio.seq.RichSequenceIterator;

import pchen.classifiers.InputOutput;
import weka.attributeSelection.WrapperSubsetEval;
import weka.classifiers.Classifier;
import weka.classifiers.bayes.BayesNet;
import weka.classifiers.bayes.NaiveBayes;
import weka.classifiers.bayes.NaiveBayesMultinomial;
import weka.classifiers.functions.LibLINEAR;
import weka.classifiers.functions.LibSVM;
import weka.classifiers.functions.RBFNetwork;
import weka.classifiers.functions.SMO;
import weka.classifiers.functions.VotedPerceptron;
import weka.classifiers.lazy.IBk;
import weka.classifiers.lazy.KStar;
import weka.classifiers.lazy.LWL;
import weka.classifiers.rules.DecisionTable;
import weka.classifiers.rules.JRip;
import weka.classifiers.rules.OneR;
import weka.classifiers.rules.PART;
import weka.classifiers.trees.J48;
import weka.classifiers.trees.RandomForest;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.SelectedTag;
import weka.core.Utils;

public class Predictor {
	private static final int winLen = 11;
	static String inputSeqAdd = ".\\Results\\Hotspots\\test\\FASTA.seq";
	static String outputSaveAdd = ".\\Results\\Hotspots\\";
	static String pdbAddr = ".\\Results\\Hotspots\\test\\PDB";
	static String AAindex1FileAddr = "C:\\Users\\chenp\\Work\\Dataset\\AAindex\\aaindex1";

	public static void main(String[] args) throws Exception {
		// prepareData();
		File dir = new File(".\\Results\\Hotspots\\train\\" + "SP\\");
		File[] trainFiles = dir.listFiles();
		dir = new File(".\\Results\\Hotspots\\test\\" + "SP\\");
		File[] testFiles = dir.listFiles();
		double[][] AAindex1 = EncoderData.getIrrelevanceAAindex2(
				AAindex1FileAddr, 0.7);
		EvalPerform4Matlab(trainFiles, testFiles, AAindex1);
	}

	/**
	 * Evaluate classifiers for the top efficient feature subsets
	 * 
	 * @param null
	 * @return null
	 * @throws FileNotFoundException
	 * @throws BioException
	 * @throws NoSuchElementException
	 * 
	 * @method Get all required data for the input of our methods
	 */
	private static void prepareData() throws FileNotFoundException,
			NoSuchElementException, BioException {
		SimpleNamespace ns = new SimpleNamespace("bj_blast");
		new EncoderData(outputSaveAdd, outputSaveAdd + "Encoder\\");

		RichSequenceIterator rs = RichSequence.IOTools.readFastaProtein(
				new BufferedReader(new FileReader(inputSeqAdd)), ns);
		while (rs.hasNext()) {
			RichSequence rr;
			rr = rs.nextRichSequence();
			Blast.BlastParser(outputSaveAdd + "Blast\\" + rr.getName()
					+ ".blast", rr.getName(), outputSaveAdd + "SP\\");
		}
	}

	/**
	 * Evaluate classifiers for the top efficient feature subsets
	 * 
	 * @param files
	 *            A File structure stores file information from a given
	 *            directory.
	 * @param testFiles
	 * @param numFolds
	 *            Set the number of folds, where one subset is for test data set
	 *            and the other for training data set.
	 * @param AAindex1
	 *            AAindex1 matrix, which row is one index composed of 20 scores
	 *            of amino acid.
	 * @param numClassifiers
	 *            Set the number of classifiers that consists of balanced pair
	 *            of positive subset and negative subset.
	 * @param topFeature
	 *            Set the number of top feature subsets.
	 * @return null
	 * @throws Exception
	 * @method First get the top feature subsets, then construct several
	 *         classifiers each one consists of balanced pair of positive subset
	 *         and negative subset, run these classifiers, combine their results
	 *         by majority voting.
	 * @throws IllegalArgumentException
	 */
	private static void EvalPerform4Matlab(File[] trainFiles, File[] testFiles,
			double[][] AAindex1) throws Exception {
		int[] trainSet = new int[trainFiles.length];
		int[] testSet = new int[testFiles.length];
		int numAAindex1 = AAindex1.length;
		for (int numTrain = 0; numTrain < trainFiles.length; numTrain++)
			trainSet[numTrain] = numTrain;
		for (int numTest = 0; numTest < testFiles.length; numTest++)
			testSet[numTest] = numTest;
		// construct classifiers having number of noNegative
		double[] informGain = new double[numAAindex1];
		getTrainInstances("train", trainSet, trainFiles, AAindex1);
		getTrainInstances("test", testSet, testFiles, AAindex1);
		Instances train0 = InputOutput
				.evaluateData(".\\Results\\Outcomes\\train" + winLen + ".arff");
		Instances train = resampleBalanced(train0);

		String[] classSym = { "BayesNet", "DecisionTable", "IBk", "J48",
				"JRip", "KStar", "LWL", "LibLINEAR", "NaiveBayes",
				"NaiveBayesMultinomial", "OneR", "PART", "RBFNetwork",
				"RandomForest", "SMO", "VotedPerceptron", "LibSVM" };
		Classifier[] clf = { new BayesNet(), new DecisionTable(), new IBk(),
				new J48(), new JRip(), new KStar(), new LWL(), new LibLINEAR(),
				new NaiveBayes(), new NaiveBayesMultinomial(), new OneR(),
				new PART(), new RBFNetwork(), new RandomForest(), new SMO(),
				new VotedPerceptron(), getLibSVM() };
		for (int classifier = 2; classifier < 3; classifier++) {
			// Rank the importance of features
			for (int noAAindex1 = 0; noAAindex1 < numAAindex1; noAAindex1++) {
				BitSet subset = new BitSet();
				subset.set((int) (noAAindex1 * winLen), (int) ((noAAindex1 + 1)
						* winLen - 1));
				informGain[noAAindex1] = FeatureSelection(train, subset,
						clf[classifier]);
			}

			Instances Test = InputOutput
					.evaluateData(".\\Results\\Outcomes\\test" + winLen
							+ ".arff");
			int numResiduePairs = Test.numInstances();
			String[] title = new String[numAAindex1 + 1];
			double[][] preds = new double[numResiduePairs][numAAindex1 + 1];
			double[][] probPredictionROC = new double[numResiduePairs][numAAindex1 + 1];
			double[] testClasses = new double[numResiduePairs];
			for (int topFeature2 = 0; topFeature2 < numAAindex1; topFeature2++) {
				System.out.println("********topFeature:" + topFeature2
						+ "********");
				Classifier testClassifier0[] = new Classifier[topFeature2 + 1];
				Test = InputOutput.evaluateData(".\\Results\\Outcomes\\test"
						+ winLen + ".arff");

				double[] tempIG = informGain.clone();
				int[] rankIG0 = Utils.sort(tempIG);
				int[] rankIG = rankIG0.clone();
				for (int i = 0; i < rankIG.length; i++)
					rankIG[i] = rankIG0[rankIG.length - 1 - i];
				int[] delIG = new int[numAAindex1 - topFeature2];
				for (int noIG = 0; noIG < numAAindex1 - topFeature2; noIG++)
					delIG[noIG] = rankIG[noIG];

				Instances emptyTrain = new Instances(train);
				int numAttr = emptyTrain.numAttributes();
				for (int noIG = 0; noIG < numAttr - (winLen + 1); noIG++)
					emptyTrain.deleteAttributeAt(0);
				Instances topTrain = emptyTrain;
				for (int noIG = topFeature2; noIG <= topFeature2; noIG++) {
					for (int noIns = 0; noIns < train.numInstances(); noIns++) {
						Instance temp0 = new DenseInstance(winLen + 1);
						for (int windowLen = 0; windowLen < winLen; windowLen++)
							temp0.setValue(windowLen, train.get(noIns).value(
									rankIG0[noIG] * winLen + windowLen));
						temp0.setValue(winLen, train.get(noIns).classValue());
						topTrain.add(noIns, temp0);
					}
				}
				// topTrain = Normalization.Scale(topTrain);

				Instances emptyTest = new Instances(Test);
				for (int noIG = 0; noIG < numAttr - (winLen + 1); noIG++)
					emptyTest.deleteAttributeAt(0);
				Instances topTest = emptyTest;
				for (int noIG = topFeature2; noIG <= topFeature2; noIG++) {
					for (int noIns = 0; noIns < Test.numInstances(); noIns++) {
						Instance temp0 = new DenseInstance(winLen + 1);
						for (int windowLen = 0; windowLen < winLen; windowLen++)
							temp0.setValue(windowLen, Test.get(noIns).value(
									rankIG0[noIG] * winLen + windowLen));
						temp0.setValue(winLen, Test.get(noIns).classValue());
						topTest.add(noIns, temp0);
					}
				}
				// topTest = Normalization.Scale(topTest);

				testClassifier0[topFeature2] = clf[classifier];
				testClassifier0[topFeature2].buildClassifier(topTrain);
				for (int m = 0; m < numResiduePairs; m++) {
					Instance instance = topTest.get(m);
					testClasses[m] = instance
							.value(topTest.numAttributes() - 1);
					preds[m][topFeature2] = testClassifier0[topFeature2]
							.classifyInstance(instance);// prediction
					probPredictionROC[m][topFeature2] = testClassifier0[topFeature2]
							.distributionForInstance(instance)[0]; // probability
					preds[m][preds[0].length - 1] = instance.value(topTest
							.numAttributes() - 1);
					probPredictionROC[m][preds[0].length - 1] = instance
							.value(topTest.numAttributes() - 1);
				}
				title[topFeature2] = "F" + topFeature2;
			}

			title[title.length - 1] = "class";
			InputOutput.saveArrayToCSV(".\\Results\\Outcomes\\ProbEachFeature"
					+ classSym[classifier] + winLen + ".csv", title,
					probPredictionROC);
		}
	}

	/**
	 * Feature selection of a subset of variables
	 * 
	 * @param data
	 *            Instances for evaluating a subset of variables in a group of
	 *            variables.
	 * @param subset
	 *            A subset of variables given the range of position in the group
	 *            of variables, such as [23 45].
	 * @return Information gain of the subset of variables over the whole
	 *         training instances.
	 * @method Using "WrapperSubsetEval" or "CfsSubsetEval" to evaluate
	 *         information gain of the subset of variables.
	 * @throws IllegalArgumentException
	 */
	private static double FeatureSelection(Instances data, BitSet subset,
			Classifier str) throws Exception {
		WrapperSubsetEval eval = new WrapperSubsetEval();
		eval.buildEvaluator(data);
		eval.setClassifier(str);
		double evalSubset = eval.evaluateSubset(subset);
		// System.out.println(evalSubset + " ");

		return evalSubset;
	}

	/**
	 * Creates a new dataset consistng of all the positive instances and a
	 * negative subset having the same size of the positive subset using random
	 * sampling with replacement.
	 * 
	 * @param data
	 *            a set of instances
	 * @return the new dataset
	 */
	public static Instances resampleBalanced(Instances data) {

		Instances newDataPositive = new Instances(data);
		Instances newDataNegative = new Instances(data);
		Instances dataNegative = new Instances(data);
		for (int i = 0; i < data.numInstances(); i++) {
			newDataPositive.delete(0);
			newDataNegative.delete(0);
			dataNegative.delete(0);
		}
		for (int i = 0; i < data.numInstances(); i++) {
			if (data.get(i).classValue() == 1) {
				newDataPositive.add(data.get(i));
			} else if (data.get(i).classValue() == 0) {
				newDataNegative.add(data.get(i));
			}
		}
		int numPositive = newDataPositive.numInstances();
		Random random = new Random();
		while (dataNegative.numInstances() < numPositive) {
			dataNegative.add(newDataNegative.get(random.nextInt(numPositive)));
			newDataPositive.add(newDataNegative
					.get(random.nextInt(numPositive)));
		}
		return newDataPositive;
	}

	/**
	 * Get balanced train instances with a pair of positive subset and negative
	 * subset in terms of parameter "noNegative" and save it into a text file.
	 * Each instance consists of production of sequence profiles of residue
	 * windows to all the AAindexes, where residue windows are composed of a
	 * pair of residue windows with 9 residues and a center window with 5
	 * residues. The central residues of the window pair are for the contact
	 * prediction and the central residue of the center window is located at the
	 * center between the two formers in sequence.
	 * 
	 * @param trainSet
	 *            A subset of file numbers that are from a given directory.
	 * @param files
	 *            A File structure stores file information from a given
	 *            directory.
	 * @param AAindex1
	 *            AAindex1 information, which row denotes for 20 indexes of
	 *            amino acids.
	 * @return null
	 * @throws FileNotFoundException
	 * @throws BioException
	 * @throws NoSuchElementException
	 * @method Get sequence profiles and AAindex1, get probable production of
	 *         sequence profiles to AAindex1, and finally save them into an file
	 *         in an arff format.
	 * @throws IllegalArgumentException
	 */
	private static void getTrainInstances(String title, int[] trainSet,
			File[] files, double[][] AAindex1) throws FileNotFoundException,
			NoSuchElementException, BioException {
		ArrayList<double[]> inputInstances = new ArrayList<double[]>();
		ArrayList<Double> inputTargets = new ArrayList<Double>();
		// double[][] AAindex1 = EncoderData.AAindex1Info(AAindex1FileAddr);
		double[] AAindexEach = new double[20];
		SimpleNamespace ns = new SimpleNamespace("bj_blast");
		RichSequenceIterator rs;

		for (int indexTrain = 0; indexTrain < trainSet.length; indexTrain++) {
			int index = trainSet[indexTrain];
			String PDBName = files[index].getName().substring(0, 5);
			// System.out.print(PDBName + "-" + indexTrain + " ");

			RichSequence rr = null;// get sequence
			rs = RichSequence.IOTools.readFastaProtein(new BufferedReader(
					new FileReader(outputSaveAdd + title + "\\Encoder\\"
							+ PDBName + "\\Seq.txt")), ns);
			while (rs.hasNext())
				rr = rs.nextRichSequence();

			double[][] seqProfile = InputOutput.readMultiTxt(outputSaveAdd
					+ title + "\\SP\\" + PDBName + ".sp");
			double[] targetOriginal = InputOutput.readSimpleTxt(outputSaveAdd
					+ title + "\\Encoder\\" + PDBName + "\\Target.txt");
			// System.out.println(PDBName + " " + seqProfile.length + " "
			// + seqProfile[0].length + " " + targetOriginal.length);
			double[] target = EncoderData.Target(targetOriginal);
			double[][] input = new double[target.length][winLen
					* AAindex1.length];
			for (int i = 0; i < AAindex1.length; i++) {
				for (int j = 0; j < AAindex1[0].length; j++) {
					if (Double.isNaN((double) AAindex1[i][j]))
						AAindexEach[j] = 0;
					else {
						AAindexEach[j] = AAindex1[i][j];
						// Utils.normalize(AAindexEach);
					}
				}
				double[][] input0 = EncoderData.EncoderWindow(rr.seqString(),
						targetOriginal, seqProfile, winLen, AAindexEach, title);

				for (int k = 0; k < input0.length; k++)
					for (int l = 0; l < input0[0].length; l++)
						input[k][i * winLen + l] = input0[k][l];
			}

			// InputOutput.saveArray(".\\src\\train0.arff", input);
			// System.out.println(target.length);
			for (int i = 0; i < target.length; i++) {
				double[] temp = new double[input[0].length];
				for (int l = 0; l < input[0].length; l++)
					temp[l] = input[i][l];
				// normalize each AAindex1 subset in each instance
				if (Utils.eq(Utils.sum(temp), 0) == true)
					inputInstances.add(temp);
				else if (Double.isNaN(Utils.sum(temp))) {
					inputInstances.add(new double[input[0].length]);
				} else {
					// Utils.normalize(temp);
					inputInstances.add(temp);
				}
				inputTargets.add(target[i]);
			}
		}

		double[][] dataBalanced = new double[inputTargets.size()][winLen
				* AAindex1.length + 1];
		for (int i = 0; i < dataBalanced.length; i++) {
			double[] temp0 = inputInstances.get(i);
			for (int j = 0; j < dataBalanced[0].length - 1; j++)
				dataBalanced[i][j] = temp0[j];

			dataBalanced[i][dataBalanced[0].length - 1] = inputTargets.get(i);
		}

		// get the balanced data and save them
		// double[][] dataBalanced = DataBalance(Instances, Targets, 0);
		InputOutput.saveArray(".\\Results\\Outcomes\\" + title + winLen
				+ ".arff", dataBalanced);
	}

}
