/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 */

/*
 *    InputOutput.java
 *    Copyright (C) 2012 Peng Chen, King Abdullah University of Science and 
 *    Technology, King of Saudi Arabia
 *
 */

package pchen.classifiers;

import java.io.DataOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Scanner;

import weka.classifiers.Classifier;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.converters.ArffLoader;
import weka.core.converters.CSVSaver;
import weka.core.converters.MatlabSaver;
import weka.filters.unsupervised.attribute.NumericToNominal;

public class InputOutput {
	public InputOutput() {

	}

	/**
	 * Obtain training data from an arff text file
	 * 
	 * @param fileFullName
	 *            File address for arff training file.
	 * @return data Instances for training file.
	 * @method
	 * @throws
	 */
	public static Instances evaluateData(String fileFullName) throws Exception {
		File trainingFile = new File(fileFullName);// Training file
		ArffLoader arffLoader = new ArffLoader();
		arffLoader.setSource(trainingFile);
		Instances data = arffLoader.getDataSet();
		data.setClassIndex(data.numAttributes() - 1);
		return data;
	}

	/**
	 * Convert instances to 2D array
	 * 
	 * @param train
	 *            Training instances.
	 * @return dataArray 2D data array.
	 * @method
	 * @throws
	 */
	public static double[][] instances2Array(Instances train) {
		double[][] dataArray = new double[train.numInstances()][train
				.numAttributes()];
		for (int i = 0; i < dataArray.length; i++) {
			Instance oneInstance = train.get(i);
			for (int j = 0; j < dataArray[0].length; j++) {
				dataArray[i][j] = oneInstance.value(j);
			}
		}
		return dataArray;
	}

	/**
	 * Save classifier model into a file
	 * 
	 * @param fileFullName
	 *            File address to save.
	 * @param testClassifier
	 *            Classifier for testing instances.
	 * @return null
	 * @method
	 * @throws
	 */
	public void saveModel(String fileFullName, Classifier testClassifier)
			throws IOException {
		File savefile = new File(fileFullName);
		OutputStream os = new FileOutputStream(savefile);
		ObjectOutputStream objectOutputStream = new ObjectOutputStream(os);
		objectOutputStream.writeObject(testClassifier);
		os.close();
	}

	/**
	 * Convert array to instance and then save it into a file in arff format
	 * 
	 * @param fileFullName
	 *            File address to save.
	 * @param test
	 *            Test instances used to offer targets of input vectors.
	 * @param predCombiner
	 *            A 2D array.
	 * @return null
	 * @method
	 * @throws
	 */
	public void saveArray(String fileFullName, Instances test,
			double[][] predCombiner) {
		try {
			FileOutputStream writer = new FileOutputStream(fileFullName);
			DataOutputStream osw = new DataOutputStream(writer);
			/*
			 * DecimalFormat df = new DecimalFormat("0.000"); // format a double
			 * value String num = df.format(a);
			 */
			osw.write("@RELATION PPI\r\n".getBytes());
			osw.write("\r\n".getBytes());
			for (int m = 0; m < predCombiner.length; m++) {
				osw.write(("@ATTRIBUTE " + m + " REAL\r\n").getBytes());
			}
			osw.write("@ATTRIBUTE class  {0,1}\r\n".getBytes());
			osw.write("\r\n".getBytes());
			osw.write("@DATA\r\n".getBytes());

			for (int m = 0; m < predCombiner[0].length; m++) {
				for (int n = 0; n < predCombiner.length; n++) {
					osw.write((String.valueOf(predCombiner[n][m]) + ", ")
							.getBytes());
				}
				osw.write("\r\n".getBytes()); // enter
				osw.write(String
						.valueOf(
								test.attributeToDoubleArray(test
										.numAttributes() - 1)[m]).getBytes());
			}
			osw.close();
			writer.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Convert array to instance and then save it into a file in arff format
	 * 
	 * @param fileFullName
	 *            File address to save.
	 * @param target
	 *            Target for test input.
	 * @param inputVector
	 *            A 2D array for test inputs.
	 * @return null
	 * @method
	 * @throws
	 */
	public void saveArray(String fileFullName, double[] target,
			double[][] inputVector) {
		try {
			FileOutputStream writer = new FileOutputStream(fileFullName);
			DataOutputStream osw = new DataOutputStream(writer);
			/*
			 * DecimalFormat df = new DecimalFormat("0.000"); // format a double
			 * value String num = df.format(a);
			 */
			DecimalFormat df = new DecimalFormat("0.000"); // format a double
			osw.write("@RELATION PPI\r\n".getBytes());
			osw.write("\r\n".getBytes());
			for (int m = 0; m < inputVector[0].length; m++) {
				osw.write(("@ATTRIBUTE " + m + " REAL\r\n").getBytes());
			}
			osw.write("@ATTRIBUTE class  {0,1}\r\n".getBytes());
			osw.write("\r\n".getBytes());
			osw.write("@DATA\r\n".getBytes());

			for (int m = 0; m < inputVector.length; m++) {
				for (int n = 0; n < inputVector[0].length; n++) {
					osw
							.write((String
									.valueOf(df.format(inputVector[m][n])) + ", ")
									.getBytes());
				}
				osw.write(String.valueOf((int) target[m]).getBytes());
				osw.write("\r\n".getBytes()); // enter
			}
			osw.close();
			writer.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Convert array to instance and then save it into a file in arff format
	 * 
	 * @param fileFullName
	 *            File address to save.
	 * @param inputVector
	 *            A 2D array for test inputs, the last column of which is for
	 *            target.
	 * @return null
	 * @method
	 * @throws
	 */
	public static void saveArray(String fileFullName, double[][] inputVector) {
		try {
			FileOutputStream writer = new FileOutputStream(fileFullName);
			DataOutputStream osw = new DataOutputStream(writer);
			/*
			 * DecimalFormat df = new DecimalFormat("0.000"); // format a double
			 * value String num = df.format(a);
			 */
			DecimalFormat df = new DecimalFormat("0.0000"); // format a double
			osw.write("@RELATION PPI\r\n".getBytes());
			osw.write("\r\n".getBytes());
			for (int m = 0; m < inputVector[0].length - 1; m++) {
				osw.write(("@ATTRIBUTE " + m + " REAL\r\n").getBytes());
			}
			osw.write("@ATTRIBUTE class  {0,1}\r\n".getBytes());
			osw.write("\r\n".getBytes());
			osw.write("@DATA\r\n".getBytes());

			for (int m = 0; m < inputVector.length; m++) {
				for (int n = 0; n < inputVector[0].length - 1; n++) {
					osw
							.write((String
									.valueOf(df.format(inputVector[m][n])) + ", ")
									.getBytes());
				}
				osw.write(String.valueOf(
						(int) inputVector[m][inputVector[0].length - 1])
						.getBytes());
				// System.out.print(predCombiner[11][m] + " ");
				osw.write("\r\n".getBytes()); // enter
			}
			osw.close();
			writer.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public static void saveArrayToMatlab(String fileFullName, double[][] array)
			throws IOException {
		MatlabSaver saver0 = new MatlabSaver();
		ArrayList<Attribute> attributes = new ArrayList<Attribute>();
		int numAttributes = array[0].length;
		int numInstances = array.length;
		for (int j = 0; j < numAttributes; j++) {
			attributes.add(new Attribute("Att" + j));
		}

		Instances temp = new Instances("PPI", attributes, 0);
		for (int j = 0; j < numInstances; j++) {
			Instance temp0 = new DenseInstance(numAttributes);
			for (int k = 0; k < numAttributes; k++) {
				temp0.setValue(k, array[j][k]);
			}
			temp.add(j, temp0);
		}
		saver0.setInstances(temp);
		saver0.setFile(new File(fileFullName));
		saver0.setUseDouble(false);
		saver0.setUseTabs(false);
		saver0.writeBatch();
	}

	public static Instances ArrayToInstances(double[][] array)
			throws IOException {
		ArrayList<Attribute> attributes = new ArrayList<Attribute>();
		int numAttributes = array[0].length;
		int numInstances = array.length;
		for (int j = 0; j < numAttributes; j++) {
			attributes.add(new Attribute("Att" + j));
		}

		Instances temp = new Instances("PPI", attributes, 0);
		for (int j = 0; j < numInstances; j++) {
			Instance temp0 = new DenseInstance(numAttributes);
			for (int k = 0; k < numAttributes; k++) {
				temp0.setValue(k, array[j][k]);
			}
			temp.add(j, temp0);
		}
		return temp;
	}

	public static void saveArrayToCSV(String fileFullName, String[] title,
			double[][] array) throws IOException {
		CSVSaver saver0 = new CSVSaver();
		ArrayList<Attribute> attributes = new ArrayList<Attribute>();
		int numAttributes = array[0].length;
		int numInstances = array.length;
		for (int j = 0; j < numAttributes; j++) {
			attributes.add(new Attribute(title[j]));
		}

		Instances temp = new Instances("PPI", attributes, 0);
		for (int j = 0; j < numInstances; j++) {
			Instance temp0 = new DenseInstance(numAttributes);
			for (int k = 0; k < numAttributes; k++) {
				temp0.setValue(k, Math.floor(array[j][k] * 1000 + .5) / 1000);
			}
			temp.add(j, temp0);
		}
		saver0.setInstances(temp);
		saver0.setFile(new File(fileFullName));
		saver0.writeBatch();
	}

	public static Instances arrayToInstances(String[] attributesType,
			double[][] array) throws Exception {
		int INIT_CAPACITY = array.length;
		ArrayList<Attribute> attributes = new ArrayList<Attribute>();
		attributesType[attributesType.length - 1] = "class  {0,1}";

		for (int i = 0; i < attributesType.length; i++) {
			attributes.add(new Attribute(attributesType[i]));
		}
		Instances instances = new Instances("Weka", attributes, INIT_CAPACITY);
		for (int i = 0; i < array.length; i++) {
			Instance instance = new DenseInstance(attributesType.length);
			for (int j = 0; j < array[0].length - 1; j++) {
				instance.setValue(attributes.get(j), array[i][j]);
			}
			instance.setValue(array[0].length - 1,
					array[i][array[0].length - 1]);
			instance.setDataset(instances);
			instance.setWeight(1);
			instances.add(i, instance);
		}
		instances.setClassIndex(instances.numAttributes() - 1);
		// Convert numeric target to nominal attribute
		if (instances.attribute(instances.classIndex()).isNumeric()) {
			NumericToNominal numToNom = new NumericToNominal();
			String[] options = new String[2];
			options[0] = "-R";
			int c_i = instances.classIndex() + 1;
			// System.out.println("[AbstractCallableLearningAlgo] class index +1: "+c_i);
			options[1] = Integer.toString(c_i);
			numToNom.setOptions(options);
			numToNom.setInputFormat(instances);
			numToNom.setAttributeIndices(String.valueOf(array[0].length));
			instances = NumericToNominal.useFilter(instances, numToNom);
		}
		return instances;
	}

	/**
	 * Read data from a multi-line text.
	 * 
	 * @param string
	 *            File address.
	 * @return seqProfile1 a 2D array for storing data.
	 * @method
	 * @throws FileNotFoundException
	 */
	public static double[][] readMultiTxt(String string)
			throws FileNotFoundException {
		Scanner in = new Scanner(new File(string));
		ArrayList<double[]> seqProfile = new ArrayList<double[]>();
		while (in.hasNextLine()) { // iterate thru the input source
			String line = in.nextLine();
			String[] strLine = line.trim().split(" ");
			double[] seqProfile0 = new double[strLine.length];
			for (int i = 0; i < strLine.length; i++)
				seqProfile0[i] = Double.parseDouble(strLine[i]);
			seqProfile.add(seqProfile0);
		}
		int row = seqProfile.size();
		int col = seqProfile.get(0).length;
		double[][] seqProfile1 = new double[row][col];
		for (int i = 0; i < row; i++)
			for (int j = 0; j < col; j++)
				seqProfile1[i][j] = seqProfile.get(i)[j];

		return seqProfile1;
	}

	/**
	 * Read data from a single-line text.
	 * 
	 * @param string
	 *            File address.
	 * @return targetOriginal1 a 1D array for storing data.
	 * @throws FileNotFoundException
	 * @method
	 */
	public static double[] readSimpleTxt(String string)
			throws FileNotFoundException {
		Scanner in = new Scanner(new File(string));
		ArrayList<Double> targetOriginal = new ArrayList<Double>();
		while (in.hasNextLine()) { // iterate thru the input source
			String line = in.nextLine();
			targetOriginal.add(Double.parseDouble(line.trim()));
		}
		int col = targetOriginal.size();
		double[] targetOriginal1 = new double[col];
		for (int j = 0; j < col; j++)
			targetOriginal1[j] = targetOriginal.get(j);

		return targetOriginal1;
	}

	/**
	 * Read data from a PSAIA text.
	 * 
	 * @param string
	 *            File address.
	 * @return targetOriginal1 a 1D array for storing data.
	 * @throws FileNotFoundException
	 * @method
	 */
	public static double[][] readPsaiaTxt(String string)
			throws FileNotFoundException {
		Scanner in = new Scanner(new File(string));
		ArrayList<double[]> psaia = new ArrayList<double[]>();
		while (in.hasNextLine()) { // iterate thru the input source
			String line = in.nextLine();
			String[] strLine = line.trim().split("\\s+");
			if (strLine.length==30) {
				double[] psaia0 = new double[strLine.length-8];
				for (int i = 0; i < strLine.length-8; i++)
					psaia0[i] = Double.parseDouble(strLine[i+8]);
				psaia.add(psaia0);
			}
		}
		int row = psaia.size();
		int col = psaia.get(0).length;
		double[][] psaia1 = new double[row][col];
		for (int i = 0; i < row; i++)
			for (int j = 0; j < col; j++)
				psaia1[i][j] = psaia.get(i)[j];

		return psaia1;
	}
}
