/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 */

/*
 *    Normalization.java
 *    Copyright (C) 2012 Peng Chen, King Abdullah University of Science and 
 *    Technology, King of Saudi Arabia
 *
 */

package pchen.classifiers;

import java.util.Arrays;

import weka.core.Instance;
import weka.core.Instances;
import weka.core.Utils;

public class Normalization {

	private static double upper = 1;
	private static double lower = 0;

	public Normalization() {
	}

	/**
	 * Normalize data into the default range [0 1]
	 * 
	 * @param data
	 *          Data for normalization
	 * @return Data scaled in the given range.
	 */
	public static double[] Scale(double[] data) {// default is the range [0 1]
		double[] data0 = data.clone();
		Arrays.sort(data0);
		double MaxV = data0[data0.length - 1]; // the last sorted value
		double MinV = data0[0];
		double[] scaled = new double[data0.length];
		for (int i = 0; i < data.length; i++) {
			scaled[i] = (data[i] - MinV) * (((upper - lower) / (MaxV - MinV)))
					+ lower;
		}
		return scaled;
	}

	/**
	 * Normalize data into the given range [lower upper]
	 * 
	 * @param data
	 *          Data for normalization.
	 * @param lower
	 *          Lower limitation for the normalization.
	 * @param upper
	 *          Upper limitation for the normalization.
	 * @return Data scaled in the given range.
	 */
	public double[] Scale(double[] data, double lower, double upper) {
		double[] data0 = data.clone();
		Arrays.sort(data0);
		double MaxV = data0[data0.length - 1]; // the last sorted value
		double MinV = data0[0];
		double[] scaled = new double[data0.length];
		for (int i = 0; i < data.length; i++) {
			scaled[i] = (data[i] - MinV) * (((upper - lower) / (MaxV - MinV)))
					+ lower;
		}
		return scaled;
	}

	/**
	 * Normalize instance data into the given range [0 1]
	 * 
	 * @param data
	 *          Data for normalization.
	 * @return Data scaled in the given range.
	 */
	public static Instances Scale(Instances data) {
		for (int i = 0; i < data.numInstances(); i++) {
			double[] array = new double[data.numAttributes() - 1];
			Instance arrayData = data.get(i);
			for (int j = 0; j < data.numAttributes() - 1; j++) {
				array[j] = arrayData.value(j);
			}
			
			if(Double.isNaN(Utils.sum(array))) {
				array = new double[data.numAttributes() - 1];
			}
			else if (Utils.eq(Utils.sum(array), 0) == false){
				Utils.normalize(array);
			}

			for (int j = 0; j < data.numAttributes() - 1; j++) {
				arrayData.setValue(j, array[j]);
			}
			data.set(i, arrayData);
		}
		return data;
	}

	/**
	 * Normalize instance data into the given range [lower upper]
	 * 
	 * @param data
	 *          Data for normalization.
	 * @param lower
	 *          Lower limitation for the normalization.
	 * @param upper
	 *          Upper limitation for the normalization.
	 * @return Data scaled in the given range.
	 */
	public Instances Scale(Instances data, int lower, int upper) {
		for (int i = 0; i < data.numInstances(); i++) {
			double[] array = new double[data.numAttributes() - 1];
			Instance arrayData = data.get(i);
			for (int j = 0; j < data.numAttributes() - 1; j++) {
				array[j] = arrayData.value(j);
			}
			double[] arrayScaled = Scale(array, lower, upper);
			for (int j = 0; j < data.numAttributes() - 1; j++) {
				arrayData.setValue(j, arrayScaled[j]);
			}
			data.set(i, arrayData);
		}
		return data;
	}

	public static double[][] Scale(double[][] data) {
		int numSample = data.length;
		int numVariable = data[0].length;
		double[][] scaled = new double[numSample][numVariable];
		for (int j = 0; j < numSample; j++) {
			double[] temp = new double[numVariable];
			for (int i = 0; i < numVariable; i++) {
				temp[i] = data[j][i];
			}
			if(Double.isNaN(Utils.sum(temp))) {
				temp = new double[numVariable];
			}
			else if (Utils.eq(Utils.sum(temp), 0) == false){
				Utils.normalize(temp);
			}
			temp = Scale(temp);
			for (int i = 0; i < numVariable; i++)
				scaled[j][i] = temp[i];
		}
		return scaled;
	}
}
